INSTAGRAM_FEED = """
{
    "created_time": "2015-10-10T10:30:00Z",
    "user": {
            "full_name": "Kevin S",
            "id": "3",
            "profile_picture": "http://distillery.s3.amazonaws.com/profiles/profile_3_75sq_1295574122.jpg",
            "username": "kevin"
        },
    "posts": [
        {
            "caption": "",
            "comment_count": 2,
            "created_time": "2015-01-01T11:00:00Z",
            "id": "123456789",
            "image": {
                "height": 612,
                "url": "http://distilleryimage2.ak.instagram.com/11f75f1cd9cc11e2a0fd22000aa8039a_7.jpg",
                "width": 612
            },
            "like_count": 1,
            "tags": [
                "example", "tag"
            ],
            "type": "image",
            "user": {
                "full_name": "Kevin S",
                "id": "3",
                "profile_picture": "http://distillery.s3.amazonaws.com/profiles/profile_3_75sq_1295574122.jpg",
                "username": "kevin"
            }
        },
        {
            "caption": "Inside le truc #foodtruck",
            "comment_count": 0,
            "created_time": "2015-01-01T11:00:00Z",
            "id": "123456789",
            "image": {
                "height": 612,
                "url": "http://distillery.s3.amazonaws.com/media/2011/02/02/6ea7baea55774c5e81e7e3e1f6e791a7_7.jpg",
                "width": 612
            },
            "like_count": 15,
            "tags": [
                "foodtruck"
            ],
            "type": "image",
            "user": {
                "full_name": "Kevin S",
                "id": "3",
                "profile_picture": "http://distillery.s3.amazonaws.com/profiles/profile_3_75sq_1295574122.jpg",
                "username": "kevin"
            }
        }
    ]
}
"""

TWITTER_FEED = """
[
    {
        "created_at": "Thu Apr 06 15:28:43 +0000 2017",
        "id": 850007368138018817,
        "id_str": "850007368138018817",
        "text": "RT @TwitterDev: 1/ Today we’re sharing our vision for the future of the Twitter API platform!nhttps://t.co/XweGngmxlP",
        "truncated": false,
        "entities": {
            "hashtags": [
            ],
            "symbols": [
            ],
            "user_mentions": [
                {
                    "screen_name": "TwitterDev",
                    "name": "TwitterDev",
                    "id": 2244994945,
                    "id_str": "2244994945",
                    "indices": [
                        3,
                        14
                    ]
                }
            ],
            "urls": [
                {
                    "url": "https://t.co/XweGngmxlP",
                    "expanded_url": "https://cards.twitter.com/cards/18ce53wgo4h/3xo1c",
                    "display_url": "cards.twitter.com/cards/18ce53wg…",
                    "indices": [
                        94,
                        117
                    ]
                }
            ]
        },
        "source": "<a href='http://twitter.com' rel='nofollow'>Twitter Web Client</a>",
        "in_reply_to_status_id": null,
        "in_reply_to_status_id_str": null,
        "in_reply_to_user_id": null,
        "in_reply_to_user_id_str": null,
        "in_reply_to_screen_name": null,
        "user": {
            "id": 6253282,
            "id_str": "6253282",
            "name": "Twitter API",
            "screen_name": "twitterapi",
            "location": "San Francisco, CA",
            "description": "The Real Twitter API. I tweet about API changes, service issues and happily answer questions about Twitter and our API. Don't get an answer? It's on my website.",
            "url": "http://t.co/78pYTvWfJd",
            "entities": {
                "url": {
                    "urls": [
                        {
                            "url": "http://t.co/78pYTvWfJd",
                            "expanded_url": "https://dev.twitter.com",
                            "display_url": "dev.twitter.com",
                            "indices": [
                                0,
                                22
                            ]
                        }
                    ]
                },
                "description": {
                    "urls": [
                    ]
                }
            },
            "protected": false,
            "followers_count": 6172353,
            "friends_count": 46,
            "listed_count": 13091,
            "created_at": "Wed May 23 06:01:13 +0000 2007",
            "favourites_count": 26,
            "utc_offset": -25200,
            "time_zone": "Pacific Time (US & Canada)",
            "geo_enabled": true,
            "verified": true,
            "statuses_count": 3583,
            "lang": "en",
            "contributors_enabled": false,
            "is_translator": false,
            "is_translation_enabled": false,
            "profile_background_color": "C0DEED",
            "profile_background_image_url": "http://pbs.twimg.com/profile_background_images/656927849/miyt9dpjz77sc0w3d4vj.png",
            "profile_background_image_url_https": "https://pbs.twimg.com/profile_background_images/656927849/miyt9dpjz77sc0w3d4vj.png",
            "profile_background_tile": true,
            "profile_image_url": "http://pbs.twimg.com/profile_images/2284174872/7df3h38zabcvjylnyfe3_normal.png",
            "profile_image_url_https": "https://pbs.twimg.com/profile_images/2284174872/7df3h38zabcvjylnyfe3_normal.png",
            "profile_banner_url": "https://pbs.twimg.com/profile_banners/6253282/1431474710",
            "profile_link_color": "0084B4",
            "profile_sidebar_border_color": "C0DEED",
            "profile_sidebar_fill_color": "DDEEF6",
            "profile_text_color": "333333",
            "profile_use_background_image": true,
            "has_extended_profile": false,
            "default_profile": false,
            "default_profile_image": false,
            "following": true,
            "follow_request_sent": false,
            "notifications": false,
            "translator_type": "regular"
        },
        "geo": null,
        "coordinates": null,
        "place": null,
        "contributors": null,
        "retweeted_status": {
            "created_at": "Thu Apr 06 15:24:15 +0000 2017",
            "id": 850006245121695744,
            "id_str": "850006245121695744",
            "text": "1/ Today we’re sharing our vision for the future of the Twitter API platform!nhttps://t.co/XweGngmxlP",
            "truncated": false,
            "entities": {
                "hashtags": [
                ],
                "symbols": [
                ],
                "user_mentions": [
                ],
                "urls": [
                    {
                        "url": "https://t.co/XweGngmxlP",
                        "expanded_url": "https://cards.twitter.com/cards/18ce53wgo4h/3xo1c",
                        "display_url": "cards.twitter.com/cards/18ce53wg…",
                        "indices": [
                            78,
                            101
                        ]
                    }
                ]
            },
            "source": "<a href='http://twitter.com' rel='nofollow'>Twitter Web Client</a>",
            "in_reply_to_status_id": null,
            "in_reply_to_status_id_str": null,
            "in_reply_to_user_id": null,
            "in_reply_to_user_id_str": null,
            "in_reply_to_screen_name": null,
            "user": {
                "id": 2244994945,
                "id_str": "2244994945",
                "name": "TwitterDev",
                "screen_name": "TwitterDev",
                "location": "Internet",
                "description": "Your official source for Twitter Platform news, updates & events. Need technical help? Visit https://t.co/mGHnxZCxkt ⌨️  #TapIntoTwitter",
                "url": "https://t.co/66w26cua1O",
                "entities": {
                    "url": {
                        "urls": [
                            {
                                "url": "https://t.co/66w26cua1O",
                                "expanded_url": "https://dev.twitter.com/",
                                "display_url": "dev.twitter.com",
                                "indices": [
                                    0,
                                    23
                                ]
                            }
                        ]
                    },
                    "description": {
                        "urls": [
                            {
                                "url": "https://t.co/mGHnxZCxkt",
                                "expanded_url": "https://twittercommunity.com/",
                                "display_url": "twittercommunity.com",
                                "indices": [
                                    93,
                                    116
                                ]
                            }
                        ]
                    }
                },
                "protected": false,
                "followers_count": 465425,
                "friends_count": 1523,
                "listed_count": 1168,
                "created_at": "Sat Dec 14 04:35:55 +0000 2013",
                "favourites_count": 2098,
                "utc_offset": -25200,
                "time_zone": "Pacific Time (US & Canada)",
                "geo_enabled": true,
                "verified": true,
                "statuses_count": 3031,
                "lang": "en",
                "contributors_enabled": false,
                "is_translator": false,
                "is_translation_enabled": false,
                "profile_background_color": "FFFFFF",
                "profile_background_image_url": "http://abs.twimg.com/images/themes/theme1/bg.png",
                "profile_background_image_url_https": "https://abs.twimg.com/images/themes/theme1/bg.png",
                "profile_background_tile": false,
                "profile_image_url": "http://pbs.twimg.com/profile_images/530814764687949824/npQQVkq8_normal.png",
                "profile_image_url_https": "https://pbs.twimg.com/profile_images/530814764687949824/npQQVkq8_normal.png",
                "profile_banner_url": "https://pbs.twimg.com/profile_banners/2244994945/1396995246",
                "profile_link_color": "0084B4",
                "profile_sidebar_border_color": "FFFFFF",
                "profile_sidebar_fill_color": "DDEEF6",
                "profile_text_color": "333333",
                "profile_use_background_image": false,
                "has_extended_profile": false,
                "default_profile": false,
                "default_profile_image": false,
                "following": true,
                "follow_request_sent": false,
                "notifications": false,
                "translator_type": "regular"
            },
            "geo": null,
            "coordinates": null,
            "place": null,
            "contributors": null,
            "is_quote_status": false,
            "retweet_count": 284,
            "favorite_count": 399,
            "favorited": false,
            "retweeted": false,
            "possibly_sensitive": false,
            "lang": "en"
        },
        "is_quote_status": false,
        "retweet_count": 284,
        "favorite_count": 0,
        "favorited": false,
        "retweeted": false,
        "possibly_sensitive": false,
        "lang": "en"
    },
    {
        "created_at": "Mon Apr 03 16:09:50 +0000 2017",
        "id": 848930551989915648,
        "id_str": "848930551989915648",
        "text": "RT @TwitterMktg: Starting today, businesses can request and share locations when engaging with people in Direct Messages. https://t.co/rpYn…",
        "truncated": false,
        "entities": {
            "hashtags": [
            ],
            "symbols": [
            ],
            "user_mentions": [
                {
                    "screen_name": "TwitterMktg",
                    "name": "Twitter Marketing",
                    "id": 357750891,
                    "id_str": "357750891",
                    "indices": [
                        3,
                        15
                    ]
                }
            ],
            "urls": [
            ]
        },
        "source": "<a href='http://twitter.com' rel='nofollow'>Twitter Web Client</a>",
        "in_reply_to_status_id": null,
        "in_reply_to_status_id_str": null,
        "in_reply_to_user_id": null,
        "in_reply_to_user_id_str": null,
        "in_reply_to_screen_name": null,
        "user": {
            "id": 6253282,
            "id_str": "6253282",
            "name": "Twitter API",
            "screen_name": "twitterapi",
            "location": "San Francisco, CA",
            "description": "The Real Twitter API. I tweet about API changes, service issues and happily answer questions about Twitter and our API. Don't get an answer? It's on my website.",
            "url": "http://t.co/78pYTvWfJd",
            "entities": {
                "url": {
                    "urls": [
                        {
                            "url": "http://t.co/78pYTvWfJd",
                            "expanded_url": "https://dev.twitter.com",
                            "display_url": "dev.twitter.com",
                            "indices": [
                                0,
                                22
                            ]
                        }
                    ]
                },
                "description": {
                    "urls": [

                    ]
                }
            },
            "protected": false,
            "followers_count": 6172353,
            "friends_count": 46,
            "listed_count": 13091,
            "created_at": "Wed May 23 06:01:13 +0000 2007",
            "favourites_count": 26,
            "utc_offset": -25200,
            "time_zone": "Pacific Time (US & Canada)",
            "geo_enabled": true,
            "verified": true,
            "statuses_count": 3583,
            "lang": "en",
            "contributors_enabled": false,
            "is_translator": false,
            "is_translation_enabled": false,
            "profile_background_color": "C0DEED",
            "profile_background_image_url": "http://pbs.twimg.com/profile_background_images/656927849/miyt9dpjz77sc0w3d4vj.png",
            "profile_background_image_url_https": "https://pbs.twimg.com/profile_background_images/656927849/miyt9dpjz77sc0w3d4vj.png",
            "profile_background_tile": true,
            "profile_image_url": "http://pbs.twimg.com/profile_images/2284174872/7df3h38zabcvjylnyfe3_normal.png",
            "profile_image_url_https": "https://pbs.twimg.com/profile_images/2284174872/7df3h38zabcvjylnyfe3_normal.png",
            "profile_banner_url": "https://pbs.twimg.com/profile_banners/6253282/1431474710",
            "profile_link_color": "0084B4",
            "profile_sidebar_border_color": "C0DEED",
            "profile_sidebar_fill_color": "DDEEF6",
            "profile_text_color": "333333",
            "profile_use_background_image": true,
            "has_extended_profile": false,
            "default_profile": false,
            "default_profile_image": false,
            "following": true,
            "follow_request_sent": false,
            "notifications": false,
            "translator_type": "regular"
        },
        "geo": null,
        "coordinates": null,
        "place": null,
        "contributors": null,
        "retweeted_status": {
            "created_at": "Mon Apr 03 16:05:05 +0000 2017",
            "id": 848929357519241216,
            "id_str": "848929357519241216",
            "text": "Starting today, businesses can request and share locations when engaging with people in Direct Messages. https://t.co/rpYndqWfQw",
            "truncated": false,
            "entities": {
                "hashtags": [
                ],
                "symbols": [
                ],
                "user_mentions": [
                ],
                "urls": [
                    {
                        "url": "https://t.co/rpYndqWfQw",
                        "expanded_url": "https://cards.twitter.com/cards/5wzucr/3x700",
                        "display_url": "cards.twitter.com/cards/5wzucr/3…",
                        "indices": [
                            105,
                            128
                        ]
                    }
                ]
            },
            "source": "<a href='https://ads.twitter.com' rel='nofollow'>Twitter Ads</a>",
            "in_reply_to_status_id": null,
            "in_reply_to_status_id_str": null,
            "in_reply_to_user_id": null,
            "in_reply_to_user_id_str": null,
            "in_reply_to_screen_name": null,
            "user": {
                "id": 357750891,
                "id_str": "357750891",
                "name": "Twitter Marketing",
                "screen_name": "TwitterMktg",
                "location": "Twitter HQ ",
                "description": "Twitter’s place for marketers, agencies, and creative thinkers ⭐ Bringing you insights, news, updates, and inspiration. Visit @TwitterAdsHelp for Ads support.",
                "url": "https://t.co/Tfo4moo92y",
                "entities": {
                    "url": {
                        "urls": [
                            {
                                "url": "https://t.co/Tfo4moo92y",
                                "expanded_url": "https://marketing.twitter.com",
                                "display_url": "marketing.twitter.com",
                                "indices": [
                                    0,
                                    23
                                ]
                            }
                        ]
                    },
                    "description": {
                        "urls": [

                        ]
                    }
                },
                "protected": false,
                "followers_count": 924546,
                "friends_count": 661,
                "listed_count": 3893,
                "created_at": "Thu Aug 18 21:08:15 +0000 2011",
                "favourites_count": 1934,
                "utc_offset": -25200,
                "time_zone": "Pacific Time (US & Canada)",
                "geo_enabled": true,
                "verified": true,
                "statuses_count": 6329,
                "lang": "en",
                "contributors_enabled": false,
                "is_translator": false,
                "is_translation_enabled": false,
                "profile_background_color": "C0DEED",
                "profile_background_image_url": "http://pbs.twimg.com/profile_background_images/662767273/jvmxdpdrplhxcw8yvkv2.png",
                "profile_background_image_url_https": "https://pbs.twimg.com/profile_background_images/662767273/jvmxdpdrplhxcw8yvkv2.png",
                "profile_background_tile": true,
                "profile_image_url": "http://pbs.twimg.com/profile_images/800953549697888256/UlXXL5h5_normal.jpg",
                "profile_image_url_https": "https://pbs.twimg.com/profile_images/800953549697888256/UlXXL5h5_normal.jpg",
                "profile_banner_url": "https://pbs.twimg.com/profile_banners/357750891/1487188210",
                "profile_link_color": "19CF86",
                "profile_sidebar_border_color": "FFFFFF",
                "profile_sidebar_fill_color": "DDEEF6",
                "profile_text_color": "333333",
                "profile_use_background_image": true,
                "has_extended_profile": false,
                "default_profile": false,
                "default_profile_image": false,
                "following": false,
                "follow_request_sent": false,
                "notifications": false,
                "translator_type": "none"
            },
            "geo": null,
            "coordinates": null,
            "place": null,
            "contributors": null,
            "is_quote_status": false,
            "retweet_count": 111,
            "favorite_count": 162,
            "favorited": false,
            "retweeted": false,
            "possibly_sensitive": false,
            "lang": "en"
        },
        "is_quote_status": false,
        "retweet_count": 111,
        "favorite_count": 0,
        "favorited": false,
        "retweeted": false,
        "lang": "en"
    }
]"""

RSS = b"""
<?xml version="1.0" encoding="utf-8" ?>
<feed xmlns="http://www.w3.org/2005/Atom">
  <title>Pellentesque Tortor</title>
  <link href="https://signagewidgets.net/samples/rss.xml" />
  <entry>
    <id>6</id>
    <title>Fusce Ligula Egestas Vestibulum</title>
    <content>Lorem ipsum dolor sit amet, consectetur adipiscing elit. Donec sed odio dui egestas eget quam.</content>
    <link rel="enclosure" type="image/jpeg" href="https://signagewidgets.net/samples/1.jpg" />
    <published>2021-08-30T10:45:12Z</published>
  </entry>
  <entry>
    <id>1</id>
    <title>Ligula Amet Venenatis Consectetur</title>
    <content>Donec ullamcorper nulla non metus auctor fringilla. Vestibulum id ligula porta felis euismod semper.</content>
    <link rel="enclosure" type="image/jpeg" href="https://signagewidgets.net/samples/2.jpg" />
    <published>2021-08-30T14:17:51Z</published>
  </entry>
  <entry>
    <id>10</id>
    <title>Condimentum Purus Commodo, donec ullamcorper nulla non metus auctor fringilla</title>
    <content>Aenean lacinia bibendum nulla sed consectetur. Fusce dapibus, tellus ac cursus commodo, ut fermentum massa justo sit amet risus.</content>
    <link rel="enclosure" type="image/jpeg" href="https://signagewidgets.net/samples/3.jpg" />
    <published>2021-08-30T15:20:30Z</published>
  </entry>
  <entry>
    <id>4</id>
    <title>Parturient Nibh Mattis Vestibulum</title>
    <content>Cras justo odio, dapibus ac facilisis in, egestas eget quam. Integer posuere erat a ante venenatis dapibus posuere velit aliquet.</content>
    <published>2021-08-30T15:20:30Z</published>
  </entry>
  <entry>
    <id>5</id>
    <title>Ridiculus Tellus Magna</title>
    <published>2021-08-30T15:20:30Z</published>
  </entry>
  <entry>
    <id>7</id>
    <title>Fusce Ligula Egestas Vestibulum</title>
    <link rel="enclosure" type="image/jpeg" href="https://signagewidgets.net/samples/4.jpg" />
    <published>2021-08-30T10:45:12Z</published>
  </entry>
  <entry>
    <id>8</id>
    <content>Cras justo odio, dapibus ac facilisis in, egestas eget quam. Integer posuere erat a ante venenatis dapibus posuere velit aliquet.</content>
    <link rel="enclosure" type="image/jpeg" href="https://signagewidgets.net/samples/5.jpg" />
    <published>2021-08-30T10:45:12Z</published>
  </entry>
  <entry>
    <id>9</id>
    <link rel="enclosure" type="image/jpeg" href="https://signagewidgets.net/samples/6.jpg" />
    <published>2021-08-30T10:45:12Z</published>
  </entry>
</feed>
"""

VIDEOS = [
    {
        "type": "video",
        "url": "https://signagewidgets.net/samples/1.mp4",
        "duration": 48.693,
        "width": 720,
        "height": 720,
    },
    {
        "type": "video",
        "url": "https://signagewidgets.net/samples/2.mp4",
        "duration": 54.73,
        "width": 720,
        "height": 720,
    },
    {
        "type": "video",
        "url": "https://signagewidgets.net/samples/3.mp4",
        "duration": 17.067,
        "width": 640,
        "height": 640,
    },
    {
        "type": "video",
        "url": "https://signagewidgets.net/samples/4.mp4",
        "duration": 16.05,
        "width": 1920,
        "height": 1080,
    },
    {
        "type": "video",
        "url": "https://signagewidgets.net/samples/5.mp4",
        "duration": 15.723,
        "width": 1920,
        "height": 1080,
    },
    {
        "type": "video",
        "url": "https://signagewidgets.net/samples/6.mp4",
        "duration": 27.904,
        "width": 1920,
        "height": 1080,
    },
    {
        "type": "video",
        "url": "https://signagewidgets.net/samples/7.mp4",
        "duration": 55.126,
        "width": 1080,
        "height": 1920,
    },
    {
        "type": "video",
        "url": "https://signagewidgets.net/samples/8.mp4",
        "duration": 10.07,
        "width": 640,
        "height": 640,
    },
]

PLAYBACK_INFO = """{
  "reason": {
    "type": "scheduled"
  },
  "campaign": {
    "id": "vPHxEp",
    "name": "content name",
    "duration": 20,
    "tags": [
      "content tag 1",
      "contenttag2"
    ],
    "attrs": {
      "attr 1": "value 1"
    }
  },
  "player": {
    "id": "8BCYL6A",
    "name": "player name",
    "version": "10.0.5",
    "tags": [
      "player tag 1",
      "playertag2"
    ],
    "attrs": {
      "attr 1": 1,
      "attr 2": "value 2"
    }
  }
}"""

PLAYBACK_LOOP = """{
  "ts": 1702061023000,
  "loops": [
    {
      "name": "PRIMARY",
      "start": 1702060000000,
      "content": {
        "id": "__simulator__",
        "playId": "#c444",
        "kind": "APP",
        "name": "app-name",
        "attrs": {},
        "reason": "LOOP",
        "start": 1702061023000
      }
    }
  ]
}"""

AIRPORT_DATA = """
{
  "name": "",
  "localtime": "",
  "updated_at": "",
  "departures": [
    {
      "location": "Asunci\u00f3n",
      "time": "2024-02-07T16:00:00-03:00",
      "estimated_time": "2024-02-07T16:00:00-03:00",
      "status": "Scheduled",
      "terminal": "",
      "gate": "",
      "flight": {
        "number": "1300",
        "iata": "PZ1300",
        "icao": "LAP1300",
        "carrier": {
          "name": "LATAM Paraguay",
          "iata": "PZ",
          "icao": "LAP",
          "logo": "https://content.airhex.com/content/logos/airlines_LAP_250_70_r.png?md5apikey=4b314e33ec57302492ae68c69ca23ff8",
          "tail": "https://content.airhex.com/content/logos/airlines_LAP_70_70_t.png?md5apikey=40828e9cb8074750d69b1eab5d55e89c"
        }
      },
      "code_shares": [
        {
          "number": "1300",
          "iata": "LA1300",
          "icao": "LAN1300",
          "carrier": {
            "name": "LATAM",
            "iata": "LA",
            "icao": "LAN",
            "logo": "https://content.airhex.com/content/logos/airlines_LAN_250_70_r.png?md5apikey=0c2539b5f76558beef49227bbaca0bcf",
            "tail": "https://content.airhex.com/content/logos/airlines_LAN_70_70_t.png?md5apikey=6c8fd45634159ea1f9f2b1ddc26ceaf9"
          }
        }
      ]
    },
    {
      "location": "Campo Grande",
      "time": "2024-02-07T16:40:00-03:00",
      "estimated_time": "2024-02-07T16:40:00-03:00",
      "status": "Scheduled",
      "terminal": "2",
      "gate": "216",
      "flight": {
        "number": "1422",
        "iata": "G31422",
        "icao": "GLO1422",
        "carrier": {
          "name": "Gol Transportes Aereos",
          "iata": "G3",
          "icao": "GLO",
          "logo": "https://content.airhex.com/content/logos/airlines_GLO_250_70_r.png?md5apikey=c4b6396e2accee41e79951b516bd07c5",
          "tail": "https://content.airhex.com/content/logos/airlines_GLO_70_70_t.png?md5apikey=f12d8358a2e9a0abfb5b6627fe0efe63"
        }
      },
      "code_shares": [
        {
          "number": "2371",
          "iata": "UX2371",
          "icao": "AEA2371",
          "carrier": {
            "name": "Air Europa",
            "iata": "UX",
            "icao": "AEA",
            "logo": "https://content.airhex.com/content/logos/airlines_AEA_250_70_r.png?md5apikey=fbb7dcad42fc78003b5ee23c1c1fb7cf",
            "tail": "https://content.airhex.com/content/logos/airlines_AEA_70_70_t.png?md5apikey=5649a3090f42745bd20389192c0a44c7"
          }
        },
        {
          "number": "8163",
          "iata": "AM8163",
          "icao": "AMX8163",
          "carrier": {
            "name": "AeroM\u00e9xico",
            "iata": "AM",
            "icao": "AMX",
            "logo": "https://content.airhex.com/content/logos/airlines_AMX_250_70_r.png?md5apikey=ae35df8024b0ae2214ecca252cada510",
            "tail": "https://content.airhex.com/content/logos/airlines_AMX_70_70_t.png?md5apikey=94b08956e269ba7e9f3f7446b1ce2a2d"
          }
        },
        {
          "number": "7632",
          "iata": "AR7632",
          "icao": "ARG7632",
          "carrier": {
            "name": "Aerolineas Argentinas",
            "iata": "AR",
            "icao": "ARG",
            "logo": "https://content.airhex.com/content/logos/airlines_ARG_250_70_r.png?md5apikey=ce98937e4fb946018be43b9a3c59d9c5",
            "tail": "https://content.airhex.com/content/logos/airlines_ARG_70_70_t.png?md5apikey=aa1e87abfff5120cc4bc70ca44d72cde"
          }
        },
        {
          "number": "3606",
          "iata": "CM3606",
          "icao": "CMP3606",
          "carrier": {
            "name": "Copa Airlines",
            "iata": "CM",
            "icao": "CMP",
            "logo": "https://content.airhex.com/content/logos/airlines_CMP_250_70_r.png?md5apikey=cf5a6f46832bd2a6557c2b827973cd71",
            "tail": "https://content.airhex.com/content/logos/airlines_CMP_70_70_t.png?md5apikey=9cf5fae85062bd85cf458a1a257baeb1"
          }
        },
        {
          "number": "4075",
          "iata": "TP4075",
          "icao": "TAP4075",
          "carrier": {
            "name": "TAP Air Portugal",
            "iata": "TP",
            "icao": "TAP",
            "logo": "https://content.airhex.com/content/logos/airlines_TAP_250_70_r.png?md5apikey=b1bb5bc624a196bcf264e3ecdad6ed97",
            "tail": "https://content.airhex.com/content/logos/airlines_TAP_70_70_t.png?md5apikey=bf762afdab016c637b7c049a1880f2b8"
          }
        }
      ]
    },
    {
      "location": "Rio de Janeiro-Gale\u00e3o International Airport",
      "time": "2024-02-07T16:55:00-03:00",
      "estimated_time": "2024-02-07T16:55:00-03:00",
      "status": "Scheduled",
      "terminal": "2",
      "gate": "",
      "flight": {
        "number": "3475",
        "iata": "JJ3475",
        "icao": "TAM3475",
        "carrier": {
          "name": "LATAM Brasil",
          "iata": "JJ",
          "icao": "TAM",
          "logo": "https://content.airhex.com/content/logos/airlines_TAM_250_70_r.png?md5apikey=cb0ce82c61b008e3ab6d27c019f87a0f",
          "tail": "https://content.airhex.com/content/logos/airlines_TAM_70_70_t.png?md5apikey=a3be93d8ca62e3d00f4363cbc6ef848c"
        }
      },
      "code_shares": [
        {
          "number": "8522",
          "iata": "AM8522",
          "icao": "AMX8522",
          "carrier": {
            "name": "AeroM\u00e9xico",
            "iata": "AM",
            "icao": "AMX",
            "logo": "https://content.airhex.com/content/logos/airlines_AMX_250_70_r.png?md5apikey=ae35df8024b0ae2214ecca252cada510",
            "tail": "https://content.airhex.com/content/logos/airlines_AMX_70_70_t.png?md5apikey=94b08956e269ba7e9f3f7446b1ce2a2d"
          }
        },
        {
          "number": "6716",
          "iata": "DL6716",
          "icao": "DAL6716",
          "carrier": {
            "name": "Delta",
            "iata": "DL",
            "icao": "DAL",
            "logo": "https://content.airhex.com/content/logos/airlines_DAL_250_70_r.png?md5apikey=a44c2ef322f43a505da3b6654cf6bff5",
            "tail": "https://content.airhex.com/content/logos/airlines_DAL_70_70_t.png?md5apikey=c7d4aa0c4e033883b5bf572cd2e7875c"
          }
        },
        {
          "number": "3475",
          "iata": "LA3475",
          "icao": "LAN3475",
          "carrier": {
            "name": "LATAM",
            "iata": "LA",
            "icao": "LAN",
            "logo": "https://content.airhex.com/content/logos/airlines_LAN_250_70_r.png?md5apikey=0c2539b5f76558beef49227bbaca0bcf",
            "tail": "https://content.airhex.com/content/logos/airlines_LAN_70_70_t.png?md5apikey=6c8fd45634159ea1f9f2b1ddc26ceaf9"
          }
        },
        {
          "number": "4847",
          "iata": "QR4847",
          "icao": "QTR4847",
          "carrier": {
            "name": "Qatar Airways",
            "iata": "QR",
            "icao": "QTR",
            "logo": "https://content.airhex.com/content/logos/airlines_QTR_250_70_r.png?md5apikey=1ecfadedfbe3e921e528861036ca3e3e",
            "tail": "https://content.airhex.com/content/logos/airlines_QTR_70_70_t.png?md5apikey=13b2f9bf97d345abaca5f996bf02ca9e"
          }
        },
        {
          "number": "7252",
          "iata": "QR7252",
          "icao": "QTR7252",
          "carrier": {
            "name": "Qatar Airways",
            "iata": "QR",
            "icao": "QTR",
            "logo": "https://content.airhex.com/content/logos/airlines_QTR_250_70_r.png?md5apikey=1ecfadedfbe3e921e528861036ca3e3e",
            "tail": "https://content.airhex.com/content/logos/airlines_QTR_70_70_t.png?md5apikey=13b2f9bf97d345abaca5f996bf02ca9e"
          }
        }
      ]
    },
    {
      "location": "Ribeir\u00e3o Preto",
      "time": "2024-02-07T16:55:00-03:00",
      "estimated_time": "2024-02-07T16:55:00-03:00",
      "status": "Scheduled",
      "terminal": "2",
      "gate": "202",
      "flight": {
        "number": "2215",
        "iata": "JJ2215",
        "icao": "TAM2215",
        "carrier": {
          "name": "LATAM Brasil",
          "iata": "JJ",
          "icao": "TAM",
          "logo": "https://content.airhex.com/content/logos/airlines_TAM_250_70_r.png?md5apikey=cb0ce82c61b008e3ab6d27c019f87a0f",
          "tail": "https://content.airhex.com/content/logos/airlines_TAM_70_70_t.png?md5apikey=a3be93d8ca62e3d00f4363cbc6ef848c"
        }
      },
      "code_shares": [
        {
          "number": "2215",
          "iata": "2Z2215",
          "icao": "PTB2215",
          "carrier": {
            "name": "Voepass Linhas A\u00e9reas",
            "iata": "2Z",
            "icao": "PTB",
            "logo": "https://content.airhex.com/content/logos/airlines_PTB_250_70_r.png?md5apikey=de629c273406e8aabc6271d0be1c6ab7",
            "tail": "https://content.airhex.com/content/logos/airlines_PTB_70_70_t.png?md5apikey=8c67f52e845e0dafefd59c4e8ca3aeb5"
          }
        }
      ]
    },
    {
      "location": "Vit\u00f3ria da Conquista",
      "time": "2024-02-07T17:00:00-03:00",
      "estimated_time": "2024-02-07T17:00:00-03:00",
      "status": "Scheduled",
      "terminal": "",
      "gate": "",
      "flight": {
        "number": "3238",
        "iata": "JJ3238",
        "icao": "TAM3238",
        "carrier": {
          "name": "LATAM Brasil",
          "iata": "JJ",
          "icao": "TAM",
          "logo": "https://content.airhex.com/content/logos/airlines_TAM_250_70_r.png?md5apikey=cb0ce82c61b008e3ab6d27c019f87a0f",
          "tail": "https://content.airhex.com/content/logos/airlines_TAM_70_70_t.png?md5apikey=a3be93d8ca62e3d00f4363cbc6ef848c"
        }
      },
      "code_shares": [
        {
          "number": "3238",
          "iata": "LA3238",
          "icao": "LAN3238",
          "carrier": {
            "name": "LATAM",
            "iata": "LA",
            "icao": "LAN",
            "logo": "https://content.airhex.com/content/logos/airlines_LAN_250_70_r.png?md5apikey=0c2539b5f76558beef49227bbaca0bcf",
            "tail": "https://content.airhex.com/content/logos/airlines_LAN_70_70_t.png?md5apikey=6c8fd45634159ea1f9f2b1ddc26ceaf9"
          }
        },
        {
          "number": "7290",
          "iata": "QR7290",
          "icao": "QTR7290",
          "carrier": {
            "name": "Qatar Airways",
            "iata": "QR",
            "icao": "QTR",
            "logo": "https://content.airhex.com/content/logos/airlines_QTR_250_70_r.png?md5apikey=1ecfadedfbe3e921e528861036ca3e3e",
            "tail": "https://content.airhex.com/content/logos/airlines_QTR_70_70_t.png?md5apikey=13b2f9bf97d345abaca5f996bf02ca9e"
          }
        }
      ]
    },
    {
      "location": "Natal",
      "time": "2024-02-07T17:00:00-03:00",
      "estimated_time": "2024-02-07T17:00:00-03:00",
      "status": "Scheduled",
      "terminal": "",
      "gate": "",
      "flight": {
        "number": "1370",
        "iata": "G31370",
        "icao": "GLO1370",
        "carrier": {
          "name": "Gol Transportes Aereos",
          "iata": "G3",
          "icao": "GLO",
          "logo": "https://content.airhex.com/content/logos/airlines_GLO_250_70_r.png?md5apikey=c4b6396e2accee41e79951b516bd07c5",
          "tail": "https://content.airhex.com/content/logos/airlines_GLO_70_70_t.png?md5apikey=f12d8358a2e9a0abfb5b6627fe0efe63"
        }
      },
      "code_shares": [
        {
          "number": "4645",
          "iata": "AV4645",
          "icao": "AVA4645",
          "carrier": {
            "name": "Avianca",
            "iata": "AV",
            "icao": "AVA",
            "logo": "https://content.airhex.com/content/logos/airlines_AVA_250_70_r.png?md5apikey=a4da4032c6fdbb8293a9061e352cdc1d",
            "tail": "https://content.airhex.com/content/logos/airlines_AVA_70_70_t.png?md5apikey=a061e5b9cc2f366fe9c6543198498c78"
          }
        },
        {
          "number": "3716",
          "iata": "CM3716",
          "icao": "CMP3716",
          "carrier": {
            "name": "Copa Airlines",
            "iata": "CM",
            "icao": "CMP",
            "logo": "https://content.airhex.com/content/logos/airlines_CMP_250_70_r.png?md5apikey=cf5a6f46832bd2a6557c2b827973cd71",
            "tail": "https://content.airhex.com/content/logos/airlines_CMP_70_70_t.png?md5apikey=9cf5fae85062bd85cf458a1a257baeb1"
          }
        },
        {
          "number": "5754",
          "iata": "TP5754",
          "icao": "TAP5754",
          "carrier": {
            "name": "TAP Air Portugal",
            "iata": "TP",
            "icao": "TAP",
            "logo": "https://content.airhex.com/content/logos/airlines_TAP_250_70_r.png?md5apikey=b1bb5bc624a196bcf264e3ecdad6ed97",
            "tail": "https://content.airhex.com/content/logos/airlines_TAP_70_70_t.png?md5apikey=bf762afdab016c637b7c049a1880f2b8"
          }
        }
      ]
    },
    {
      "location": "Buenos Aires - Ministro Pistarini International Airport",
      "time": "2024-02-07T17:00:00-03:00",
      "estimated_time": "2024-02-07T17:20:00-03:00",
      "status": "Scheduled",
      "terminal": "2",
      "gate": "272",
      "flight": {
        "number": "506",
        "iata": "ET506",
        "icao": "ETH506",
        "carrier": {
          "name": "Ethiopian Airlines",
          "iata": "ET",
          "icao": "ETH",
          "logo": "https://content.airhex.com/content/logos/airlines_ETH_250_70_r.png?md5apikey=3989d8672d7e0695be949c41d4e947e4",
          "tail": "https://content.airhex.com/content/logos/airlines_ETH_70_70_t.png?md5apikey=ea10182db5b1961228b04ebfb766950c"
        }
      },
      "code_shares": []
    },
    {
      "location": "Porto Seguro",
      "time": "2024-02-07T17:05:00-03:00",
      "estimated_time": "2024-02-07T17:05:00-03:00",
      "status": "Scheduled",
      "terminal": "",
      "gate": "",
      "flight": {
        "number": "3244",
        "iata": "JJ3244",
        "icao": "TAM3244",
        "carrier": {
          "name": "LATAM Brasil",
          "iata": "JJ",
          "icao": "TAM",
          "logo": "https://content.airhex.com/content/logos/airlines_TAM_250_70_r.png?md5apikey=cb0ce82c61b008e3ab6d27c019f87a0f",
          "tail": "https://content.airhex.com/content/logos/airlines_TAM_70_70_t.png?md5apikey=a3be93d8ca62e3d00f4363cbc6ef848c"
        }
      },
      "code_shares": [
        {
          "number": "3244",
          "iata": "LA3244",
          "icao": "LAN3244",
          "carrier": {
            "name": "LATAM",
            "iata": "LA",
            "icao": "LAN",
            "logo": "https://content.airhex.com/content/logos/airlines_LAN_250_70_r.png?md5apikey=0c2539b5f76558beef49227bbaca0bcf",
            "tail": "https://content.airhex.com/content/logos/airlines_LAN_70_70_t.png?md5apikey=6c8fd45634159ea1f9f2b1ddc26ceaf9"
          }
        },
        {
          "number": "5128",
          "iata": "QR5128",
          "icao": "QTR5128",
          "carrier": {
            "name": "Qatar Airways",
            "iata": "QR",
            "icao": "QTR",
            "logo": "https://content.airhex.com/content/logos/airlines_QTR_250_70_r.png?md5apikey=1ecfadedfbe3e921e528861036ca3e3e",
            "tail": "https://content.airhex.com/content/logos/airlines_QTR_70_70_t.png?md5apikey=13b2f9bf97d345abaca5f996bf02ca9e"
          }
        }
      ]
    }
  ],
  "arrivals": [
    {
      "location": "Rio de Janeiro-Gale\u00e3o International Airport",
      "time": "2024-02-07T15:10:00-03:00",
      "estimated_time": "2024-02-07T15:08:00-03:00",
      "status": "Taxiing",
      "terminal": "2",
      "gate": "",
      "flight": {
        "number": "3530",
        "iata": "JJ3530",
        "icao": "TAM3530",
        "carrier": {
          "name": "LATAM Brasil",
          "iata": "JJ",
          "icao": "TAM",
          "logo": "https://content.airhex.com/content/logos/airlines_TAM_250_70_r.png?md5apikey=cb0ce82c61b008e3ab6d27c019f87a0f",
          "tail": "https://content.airhex.com/content/logos/airlines_TAM_70_70_t.png?md5apikey=a3be93d8ca62e3d00f4363cbc6ef848c"
        }
      },
      "code_shares": [
        {
          "number": "3530",
          "iata": "LA3530",
          "icao": "LAN3530",
          "carrier": {
            "name": "LATAM",
            "iata": "LA",
            "icao": "LAN",
            "logo": "https://content.airhex.com/content/logos/airlines_LAN_250_70_r.png?md5apikey=0c2539b5f76558beef49227bbaca0bcf",
            "tail": "https://content.airhex.com/content/logos/airlines_LAN_70_70_t.png?md5apikey=6c8fd45634159ea1f9f2b1ddc26ceaf9"
          }
        },
        {
          "number": "4832",
          "iata": "QR4832",
          "icao": "QTR4832",
          "carrier": {
            "name": "Qatar Airways",
            "iata": "QR",
            "icao": "QTR",
            "logo": "https://content.airhex.com/content/logos/airlines_QTR_250_70_r.png?md5apikey=1ecfadedfbe3e921e528861036ca3e3e",
            "tail": "https://content.airhex.com/content/logos/airlines_QTR_70_70_t.png?md5apikey=13b2f9bf97d345abaca5f996bf02ca9e"
          }
        },
        {
          "number": "4836",
          "iata": "QR4836",
          "icao": "QTR4836",
          "carrier": {
            "name": "Qatar Airways",
            "iata": "QR",
            "icao": "QTR",
            "logo": "https://content.airhex.com/content/logos/airlines_QTR_250_70_r.png?md5apikey=1ecfadedfbe3e921e528861036ca3e3e",
            "tail": "https://content.airhex.com/content/logos/airlines_QTR_70_70_t.png?md5apikey=13b2f9bf97d345abaca5f996bf02ca9e"
          }
        }
      ]
    },
    {
      "location": "Ribeir\u00e3o Preto",
      "time": "2024-02-07T15:50:00-03:00",
      "estimated_time": "2024-02-07T15:50:00-03:00",
      "status": "Scheduled",
      "terminal": "2",
      "gate": "",
      "flight": {
        "number": "2214",
        "iata": "JJ2214",
        "icao": "TAM2214",
        "carrier": {
          "name": "LATAM Brasil",
          "iata": "JJ",
          "icao": "TAM",
          "logo": "https://content.airhex.com/content/logos/airlines_TAM_250_70_r.png?md5apikey=cb0ce82c61b008e3ab6d27c019f87a0f",
          "tail": "https://content.airhex.com/content/logos/airlines_TAM_70_70_t.png?md5apikey=a3be93d8ca62e3d00f4363cbc6ef848c"
        }
      },
      "code_shares": [
        {
          "number": "2214",
          "iata": "2Z2214",
          "icao": "PTB2214",
          "carrier": {
            "name": "Voepass Linhas A\u00e9reas",
            "iata": "2Z",
            "icao": "PTB",
            "logo": "https://content.airhex.com/content/logos/airlines_PTB_250_70_r.png?md5apikey=de629c273406e8aabc6271d0be1c6ab7",
            "tail": "https://content.airhex.com/content/logos/airlines_PTB_70_70_t.png?md5apikey=8c67f52e845e0dafefd59c4e8ca3aeb5"
          }
        }
      ]
    },
    {
      "location": "Lima",
      "time": "2024-02-07T16:20:00-03:00",
      "estimated_time": "2024-02-07T17:24:00-03:00",
      "status": "Delayed",
      "terminal": "3",
      "gate": "",
      "flight": {
        "number": "8040",
        "iata": "JJ8040",
        "icao": "TAM8040",
        "carrier": {
          "name": "LATAM Brasil",
          "iata": "JJ",
          "icao": "TAM",
          "logo": "https://content.airhex.com/content/logos/airlines_TAM_250_70_r.png?md5apikey=cb0ce82c61b008e3ab6d27c019f87a0f",
          "tail": "https://content.airhex.com/content/logos/airlines_TAM_70_70_t.png?md5apikey=a3be93d8ca62e3d00f4363cbc6ef848c"
        }
      },
      "code_shares": [
        {
          "number": "6124",
          "iata": "DL6124",
          "icao": "DAL6124",
          "carrier": {
            "name": "Delta",
            "iata": "DL",
            "icao": "DAL",
            "logo": "https://content.airhex.com/content/logos/airlines_DAL_250_70_r.png?md5apikey=a44c2ef322f43a505da3b6654cf6bff5",
            "tail": "https://content.airhex.com/content/logos/airlines_DAL_70_70_t.png?md5apikey=c7d4aa0c4e033883b5bf572cd2e7875c"
          }
        },
        {
          "number": "7380",
          "iata": "DL7380",
          "icao": "DAL7380",
          "carrier": {
            "name": "Delta",
            "iata": "DL",
            "icao": "DAL",
            "logo": "https://content.airhex.com/content/logos/airlines_DAL_250_70_r.png?md5apikey=a44c2ef322f43a505da3b6654cf6bff5",
            "tail": "https://content.airhex.com/content/logos/airlines_DAL_70_70_t.png?md5apikey=c7d4aa0c4e033883b5bf572cd2e7875c"
          }
        },
        {
          "number": "8040",
          "iata": "LA8040",
          "icao": "LAN8040",
          "carrier": {
            "name": "LATAM",
            "iata": "LA",
            "icao": "LAN",
            "logo": "https://content.airhex.com/content/logos/airlines_LAN_250_70_r.png?md5apikey=0c2539b5f76558beef49227bbaca0bcf",
            "tail": "https://content.airhex.com/content/logos/airlines_LAN_70_70_t.png?md5apikey=6c8fd45634159ea1f9f2b1ddc26ceaf9"
          }
        },
        {
          "number": "5209",
          "iata": "QR5209",
          "icao": "QTR5209",
          "carrier": {
            "name": "Qatar Airways",
            "iata": "QR",
            "icao": "QTR",
            "logo": "https://content.airhex.com/content/logos/airlines_QTR_250_70_r.png?md5apikey=1ecfadedfbe3e921e528861036ca3e3e",
            "tail": "https://content.airhex.com/content/logos/airlines_QTR_70_70_t.png?md5apikey=13b2f9bf97d345abaca5f996bf02ca9e"
          }
        },
        {
          "number": "9762",
          "iata": "LX9762",
          "icao": "SWR9762",
          "carrier": {
            "name": "Swiss",
            "iata": "LX",
            "icao": "SWR",
            "logo": "https://content.airhex.com/content/logos/airlines_SWR_250_70_r.png?md5apikey=ef07165baa14865d3f421bac26f65930",
            "tail": "https://content.airhex.com/content/logos/airlines_SWR_70_70_t.png?md5apikey=5d03b5a8ea5c8599fb3a1698e4bc13f2"
          }
        }
      ]
    },
    {
      "location": "Passo Fundo",
      "time": "2024-02-07T16:20:00-03:00",
      "estimated_time": "2024-02-07T16:20:00-03:00",
      "status": "Scheduled",
      "terminal": "",
      "gate": "",
      "flight": {
        "number": "3403",
        "iata": "JJ3403",
        "icao": "TAM3403",
        "carrier": {
          "name": "LATAM Brasil",
          "iata": "JJ",
          "icao": "TAM",
          "logo": "https://content.airhex.com/content/logos/airlines_TAM_250_70_r.png?md5apikey=cb0ce82c61b008e3ab6d27c019f87a0f",
          "tail": "https://content.airhex.com/content/logos/airlines_TAM_70_70_t.png?md5apikey=a3be93d8ca62e3d00f4363cbc6ef848c"
        }
      },
      "code_shares": [
        {
          "number": "3403",
          "iata": "LA3403",
          "icao": "LAN3403",
          "carrier": {
            "name": "LATAM",
            "iata": "LA",
            "icao": "LAN",
            "logo": "https://content.airhex.com/content/logos/airlines_LAN_250_70_r.png?md5apikey=0c2539b5f76558beef49227bbaca0bcf",
            "tail": "https://content.airhex.com/content/logos/airlines_LAN_70_70_t.png?md5apikey=6c8fd45634159ea1f9f2b1ddc26ceaf9"
          }
        },
        {
          "number": "5121",
          "iata": "QR5121",
          "icao": "QTR5121",
          "carrier": {
            "name": "Qatar Airways",
            "iata": "QR",
            "icao": "QTR",
            "logo": "https://content.airhex.com/content/logos/airlines_QTR_250_70_r.png?md5apikey=1ecfadedfbe3e921e528861036ca3e3e",
            "tail": "https://content.airhex.com/content/logos/airlines_QTR_70_70_t.png?md5apikey=13b2f9bf97d345abaca5f996bf02ca9e"
          }
        },
        {
          "number": "5170",
          "iata": "QR5170",
          "icao": "QTR5170",
          "carrier": {
            "name": "Qatar Airways",
            "iata": "QR",
            "icao": "QTR",
            "logo": "https://content.airhex.com/content/logos/airlines_QTR_250_70_r.png?md5apikey=1ecfadedfbe3e921e528861036ca3e3e",
            "tail": "https://content.airhex.com/content/logos/airlines_QTR_70_70_t.png?md5apikey=13b2f9bf97d345abaca5f996bf02ca9e"
          }
        }
      ]
    },
    {
      "location": "Vit\u00f3ria",
      "time": "2024-02-07T16:25:00-03:00",
      "estimated_time": "2024-02-07T16:25:00-03:00",
      "status": "Taxiing",
      "terminal": "2",
      "gate": "",
      "flight": {
        "number": "3333",
        "iata": "JJ3333",
        "icao": "TAM3333",
        "carrier": {
          "name": "LATAM Brasil",
          "iata": "JJ",
          "icao": "TAM",
          "logo": "https://content.airhex.com/content/logos/airlines_TAM_250_70_r.png?md5apikey=cb0ce82c61b008e3ab6d27c019f87a0f",
          "tail": "https://content.airhex.com/content/logos/airlines_TAM_70_70_t.png?md5apikey=a3be93d8ca62e3d00f4363cbc6ef848c"
        }
      },
      "code_shares": [
        {
          "number": "6127",
          "iata": "DL6127",
          "icao": "DAL6127",
          "carrier": {
            "name": "Delta",
            "iata": "DL",
            "icao": "DAL",
            "logo": "https://content.airhex.com/content/logos/airlines_DAL_250_70_r.png?md5apikey=a44c2ef322f43a505da3b6654cf6bff5",
            "tail": "https://content.airhex.com/content/logos/airlines_DAL_70_70_t.png?md5apikey=c7d4aa0c4e033883b5bf572cd2e7875c"
          }
        },
        {
          "number": "6276",
          "iata": "DL6276",
          "icao": "DAL6276",
          "carrier": {
            "name": "Delta",
            "iata": "DL",
            "icao": "DAL",
            "logo": "https://content.airhex.com/content/logos/airlines_DAL_250_70_r.png?md5apikey=a44c2ef322f43a505da3b6654cf6bff5",
            "tail": "https://content.airhex.com/content/logos/airlines_DAL_70_70_t.png?md5apikey=c7d4aa0c4e033883b5bf572cd2e7875c"
          }
        },
        {
          "number": "4705",
          "iata": "LH4705",
          "icao": "DLH4705",
          "carrier": {
            "name": "Lufthansa",
            "iata": "LH",
            "icao": "DLH",
            "logo": "https://content.airhex.com/content/logos/airlines_DLH_250_70_r.png?md5apikey=3c386bbd5275e5997bcdd6434bd7e163",
            "tail": "https://content.airhex.com/content/logos/airlines_DLH_70_70_t.png?md5apikey=922e2a45ef5234356efb70991da3853c"
          }
        },
        {
          "number": "3333",
          "iata": "LA3333",
          "icao": "LAN3333",
          "carrier": {
            "name": "LATAM",
            "iata": "LA",
            "icao": "LAN",
            "logo": "https://content.airhex.com/content/logos/airlines_LAN_250_70_r.png?md5apikey=0c2539b5f76558beef49227bbaca0bcf",
            "tail": "https://content.airhex.com/content/logos/airlines_LAN_70_70_t.png?md5apikey=6c8fd45634159ea1f9f2b1ddc26ceaf9"
          }
        },
        {
          "number": "5253",
          "iata": "QR5253",
          "icao": "QTR5253",
          "carrier": {
            "name": "Qatar Airways",
            "iata": "QR",
            "icao": "QTR",
            "logo": "https://content.airhex.com/content/logos/airlines_QTR_250_70_r.png?md5apikey=1ecfadedfbe3e921e528861036ca3e3e",
            "tail": "https://content.airhex.com/content/logos/airlines_QTR_70_70_t.png?md5apikey=13b2f9bf97d345abaca5f996bf02ca9e"
          }
        }
      ]
    },
    {
      "location": "Jo\u00e3o Pessoa",
      "time": "2024-02-07T16:50:00-03:00",
      "estimated_time": "2024-02-07T17:28:17-03:00",
      "status": "Delayed",
      "terminal": "2",
      "gate": "",
      "flight": {
        "number": "1639",
        "iata": "G31639",
        "icao": "GLO1639",
        "carrier": {
          "name": "Gol Transportes Aereos",
          "iata": "G3",
          "icao": "GLO",
          "logo": "https://content.airhex.com/content/logos/airlines_GLO_250_70_r.png?md5apikey=c4b6396e2accee41e79951b516bd07c5",
          "tail": "https://content.airhex.com/content/logos/airlines_GLO_70_70_t.png?md5apikey=f12d8358a2e9a0abfb5b6627fe0efe63"
        }
      },
      "code_shares": [
        {
          "number": "3725",
          "iata": "CM3725",
          "icao": "CMP3725",
          "carrier": {
            "name": "Copa Airlines",
            "iata": "CM",
            "icao": "CMP",
            "logo": "https://content.airhex.com/content/logos/airlines_CMP_250_70_r.png?md5apikey=cf5a6f46832bd2a6557c2b827973cd71",
            "tail": "https://content.airhex.com/content/logos/airlines_CMP_70_70_t.png?md5apikey=9cf5fae85062bd85cf458a1a257baeb1"
          }
        },
        {
          "number": "9341",
          "iata": "KL9341",
          "icao": "KLM9341",
          "carrier": {
            "name": "KLM",
            "iata": "KL",
            "icao": "KLM",
            "logo": "https://content.airhex.com/content/logos/airlines_KLM_250_70_r.png?md5apikey=a315bf03b89d64d85db386f7a47ba544",
            "tail": "https://content.airhex.com/content/logos/airlines_KLM_70_70_t.png?md5apikey=08ffa80a7c3efba62650c4616bde8dc1"
          }
        },
        {
          "number": "5764",
          "iata": "TP5764",
          "icao": "TAP5764",
          "carrier": {
            "name": "TAP Air Portugal",
            "iata": "TP",
            "icao": "TAP",
            "logo": "https://content.airhex.com/content/logos/airlines_TAP_250_70_r.png?md5apikey=b1bb5bc624a196bcf264e3ecdad6ed97",
            "tail": "https://content.airhex.com/content/logos/airlines_TAP_70_70_t.png?md5apikey=bf762afdab016c637b7c049a1880f2b8"
          }
        }
      ]
    },
    {
      "location": "Cochabamba",
      "time": "2024-02-07T17:25:00-03:00",
      "estimated_time": "2024-02-07T18:20:00-03:00",
      "status": "Delayed",
      "terminal": "2",
      "gate": "",
      "flight": {
        "number": "738",
        "iata": "OB738",
        "icao": "BOV738",
        "carrier": {
          "name": "Boliviana de Aviaci\u00f3n",
          "iata": "OB",
          "icao": "BOV",
          "logo": "https://content.airhex.com/content/logos/airlines_BOV_250_70_r.png?md5apikey=82418c28de871c666b5baade66fc479d",
          "tail": "https://content.airhex.com/content/logos/airlines_BOV_70_70_t.png?md5apikey=321258b35093b4138327265a2c710d56"
        }
      },
      "code_shares": []
    },
    {
      "location": "Dubai International Airport",
      "time": "2024-02-07T17:35:00-03:00",
      "estimated_time": "2024-02-07T18:05:00-03:00",
      "status": "Delayed",
      "terminal": "3",
      "gate": "",
      "flight": {
        "number": "261",
        "iata": "EK261",
        "icao": "UAE261",
        "carrier": {
          "name": "Emirates",
          "iata": "EK",
          "icao": "UAE",
          "logo": "https://content.airhex.com/content/logos/airlines_UAE_250_70_r.png?md5apikey=6e3334b9dd1a5d05ddadd6447eed4eac",
          "tail": "https://content.airhex.com/content/logos/airlines_UAE_70_70_t.png?md5apikey=d25c2e64be2667fe69a48120d512c778"
        }
      },
      "code_shares": [
        {
          "number": "5406",
          "iata": "GF5406",
          "icao": "GFA5406",
          "carrier": {
            "name": "Gulf Air",
            "iata": "GF",
            "icao": "GFA",
            "logo": "https://content.airhex.com/content/logos/airlines_GFA_250_70_r.png?md5apikey=6294e5d93acba6318e7d7a2b8006f697",
            "tail": "https://content.airhex.com/content/logos/airlines_GFA_70_70_t.png?md5apikey=f272b4d6f673d4cf1e03c27fe56d9c07"
          }
        }
      ]
    }
  ]
}
"""



import colorsys
import datetime
import decimal
import hashlib
import json
import logging
import re
import threading
import urllib
import uuid
from collections import OrderedDict
from html.parser import HTMLParser

from jinja2 import Markup, StrictUndefined, Undefined, nodes
from jinja2.exceptions import TemplateSyntaxError
from jinja2.ext import GETTEXT_FUNCTIONS, Extension, extract_from_ast
from jinja2.sandbox import ImmutableSandboxedEnvironment

from . import samples, utils
from .fields import AVAILABLE_FONTS

_local = threading.local()

logger = logging.getLogger("onsigntv.app_config")

BANNED_NAMES = {
    "category",
    "id",
    "list",
    "media",
    "name",
    "widget_id",
    "widget_uid",
    "gettext",
    "ngettext",
    "signagewidgets",
}

KNOWN_TYPES = {
    "audio",
    "audiolist",
    "airport",
    "bool",
    "choice",
    "color",
    "currency",
    "date",
    "datetime",
    "float",
    "font",
    "googlesheet",
    "googlevoice",
    "instagram",
    "image",
    "imagelist",
    "location",
    "media",
    "multichoice",
    "multicurrency",
    "number",
    "paragraph",
    "richtext",
    "stock",
    "text",
    "time",
    "twitter",
    "url",
    "video",
    "webfeed",
    "xml",
}

JS_SUPPORTED_TYPES = {
    "audio",
    "audiolist",
    "bool",
    "choice",
    "color",
    "date",
    "datetime",
    "float",
    "image",
    "imagelist",
    "media",
    "multichoice",
    "number",
    "paragraph",
    "richtext",
    "text",
    "time",
    "url",
    "video",
    "xml",
}

USER_TYPES = {"audio", "audiolist", "image", "imagelist", "media", "video"}

KNOWN_METAS = {
    "aspectratio",
    "audio",
    "audio-app",
    "caps",
    "compatibility",
    "description",
    "automation",
    "automation-app",
    "plugin",
    "plugin-app",
}

DATA_SOURCE_TYPES = {
    "boolean",
    "date",
    "datetime",
    "image",
    "integer",
    "media",
    "number",
    "text",
    "time",
    "url",
    "video",
}

ATTR_TYPES = {
    "number",
    "string",
    "numberarray",
    "stringarray",
}

ATTR_MODES = {"r", "w", "rw"}

FETCH_TEMPLATE = """
<script type="text/plain" id="%(id)s">%(content)s</script>
<script type="text/javascript">
  window.%(variable)s = document.getElementById('%(id)s').innerHTML.trim().replace(/&(lt|gt|#34|#39|amp);/g, function(match) {
    switch(match) {
      case '&lt;': return '<'; case '&gt;': return '>'; case '&#34;': return '"'; case '&#39;': return "'"; case '&amp;': return '&';
    }
  });
</script>
"""

SDK_TAG = "<!--__loadsdk__-->"
SDK_WARNING_EXPIRY_DATE = "2023-02-01"


class MediaItem:
    def __init__(self, path, name, tracker):
        self._path = path / name
        self._name = name
        self._tracker = tracker

    @property
    def name(self):
        return self._path.name

    @property
    def url(self):
        self._tracker(self._path)
        return urllib.parse.quote(self._name) + "?tracked=1"

    @property
    def width(self):
        from PIL import Image

        image = Image.open(self._path)
        return image.size[0]

    @property
    def height(self):
        from PIL import Image

        image = Image.open(self._path)
        return image.size[1]


class ConfigurableWidgetMedia:
    def __init__(self, path, tracker):
        logger.debug("register widget media: %s", path)
        self._path = path
        self._tracker = tracker
        self._subfiles = {
            re.sub(r"[^\w]", "_", subpath.stem.lower()): MediaItem(
                self._path, subpath.name, self._tracker
            )
            for subpath in self._path.iterdir()
            if subpath.is_file() and not subpath.name.startswith(".")
        }

    def __getitem__(self, name):
        if not (self._path / name).is_file():
            raise ValueError(f"File not found: {name}")

        return MediaItem(self._path, name, self._tracker)

    def __getattr__(self, key):
        if key not in self._subfiles:
            raise ValueError(f"File not found: {key}")

        return self._subfiles[key.lower()]


class LocalContextManager:
    def __init__(self):
        self.ctx = {}

    def __call__(self, **kwargs):
        self.ctx = dict(kwargs)
        return self

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.ctx = {}

    def do_gettext(self, value):
        return value

    def do_ngettext(self, singular, plural, num):
        return singular if num == 1 else plural

    def do_localize(self, value, format=None):
        return value

    def do_shim(self, name):
        if name == "events":
            return ""

        shims = {
            "i18n": "static/shim/Intl.min.js",
        }

        path = shims.get(name)
        if path is None:
            raise ValueError("Unknown shim")

        return Markup(utils.get_resource_string(path))

    def do_datafeed(
        self, *, name, label, fields, help=None, optional=False, optgroup=None
    ):
        data_source_item = self.ctx["context"].get(name)
        if data_source_item:
            return data_source_item._render(name)
        return ""

    def do_fetch_feed(self, value):
        from .fields import WebFeed, WebFeedEntry
        from .rss import process_feed

        processed_feed = process_feed(value)

        return WebFeed(
            url=value,
            title=processed_feed["title"],
            subtitle=processed_feed["subtitle"],
            entries=[WebFeedEntry(entry) for entry in processed_feed["entries"]],
        )

    def do_fetch_sheet(self, value):
        from .fields import GoogleSheet, GoogleSheetsURLField

        match = GoogleSheetsURLField.RE_SHEET_ID.match(value)
        return GoogleSheet(match.groups(1))

    def do_fetch_text(self, url, refresh=15, variable="FETCH_RESULT"):
        try:
            resp = utils.get_url(url)
            data = resp.data
        except Exception:
            data = ""

        return Markup(FETCH_TEMPLATE) % {
            "id": "v" + hashlib.sha1(data).hexdigest(),
            "content": data.decode("utf-8"),
            "variable": variable,
        }

    def do_signagewidgets(self, *args, **kwargs):
        from urllib.parse import urlencode

        url = "https://signagewidgets.net/" + "/".join(map(str, args))

        if kwargs:
            url = url + "?" + urlencode(kwargs)

        return url


def render_app_html(context, path, tracker, **kwargs):
    context = {**context, "media": ConfigurableWidgetMedia(path.parent, tracker)}

    html = None
    exceptions = None

    env = default_jinja_env()

    try:
        # First we need to convert the data into a template
        template = env.from_string(path.read_text("utf-8"))

        with env.local_ctx(tracker=tracker, context=context):
            html = template.render(context)
    except Exception as exp:
        exceptions = exp

    return html, exceptions


def default_jinja_env():
    if hasattr(_local, "env"):
        return _local.env

    env = ImmutableSandboxedEnvironment(
        autoescape=True,
        keep_trailing_newline=True,
        undefined=StrictUndefined,
        extensions=[
            ErrorExtension,
            "jinja2.ext.autoescape",
            "jinja2.ext.InternationalizationExtension",
        ],
    )
    env.local_ctx = LocalContextManager()
    env.install_gettext_callables(
        env.local_ctx.do_gettext, env.local_ctx.do_ngettext, newstyle=True
    )

    # In the default env __config__ is a no-op.
    env.globals["__config__"] = lambda *args, **kwargs: ""
    env.globals["__meta__"] = lambda *args, **kwargs: ""

    env.globals["contrast"] = do_contrast
    env.globals["darken"] = do_darken
    env.globals["lighten"] = do_lighten
    env.globals["regex_match"] = do_regex_match
    env.globals["shim"] = env.local_ctx.do_shim
    env.globals["fetch_feed"] = env.local_ctx.do_fetch_feed
    env.globals["fetch_text"] = env.local_ctx.do_fetch_text
    env.globals["fetch_sheet"] = env.local_ctx.do_fetch_sheet
    env.globals["localize"] = env.local_ctx.do_localize
    env.globals["signagewidgets"] = env.local_ctx.do_signagewidgets
    env.globals["__datasink__"] = env.local_ctx.do_datafeed
    env.globals["__datafeed__"] = env.local_ctx.do_datafeed
    env.globals["__field__"] = lambda *args, **kwargs: ""
    env.globals["__attr__"] = lambda *args, **kwargs: ""

    env.globals["__loadsdk__"] = utils.CallableString()

    env.globals["__id__"] = "__simulator__"
    env.globals["__lang__"] = "en"
    env.globals["__rtl__"] = False
    env.globals["__muted__"] = False
    env.globals["timeline_is_muted"] = False

    env.filters["json"] = do_json
    env.filters["numberfmt"] = do_numberfmt
    env.filters["slugify"] = do_slugify
    env.filters["qrcode"] = do_qrcode

    _local.env = env

    return env


def s_rgba(value):
    match = re.match(r"\s*#([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})\s*", value, re.I)
    if match:
        return tuple(int(i, 16) / 255 for i in match.groups()) + (1.0,)

    match = re.match(
        r"\s*#([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})\s*", value, re.I
    )
    if match:
        return tuple(int(i, 16) / 255 for i in match.groups())

    match = re.match(
        r"""
        rgba\(\s*
            ([01]?[0-9]?[0-9]|2[0-4][0-9]|25[0-5])\s*,\s*
            ([01]?[0-9]?[0-9]|2[0-4][0-9]|25[0-5])\s*,\s*
            ([01]?[0-9]?[0-9]|2[0-4][0-9]|25[0-5])\s*,\s*
            (\d*\.\d+)
        \s*\)
    """,
        value,
        re.X | re.I,
    )
    if match:
        return tuple(int(i) / 255 for i in match.groups()[:3]) + (
            float(match.groups()[-1]),
        )

    raise ValueError("invalid color")


def rgb_s(r, g, b, a=1.0):
    if a == 1.0:
        return "#{:02x}{:02x}{:02x}".format(*tuple(round(c * 255) for c in (r, g, b)))
    else:
        return "rgba(%i,%i,%i,%.3f)" % tuple([round(c * 255) for c in (r, g, b)] + [a])


def s_amount(amount):
    if isinstance(amount, str):
        return float(amount.strip(" %")) / 100

    return amount


def do_lighten(value, amount=0.1):
    r, g, b, a = s_rgba(value)

    h, l, s = colorsys.rgb_to_hls(r, g, b)
    l = min(1, max(0, l + s_amount(amount)))

    return rgb_s(*colorsys.hls_to_rgb(h, l, s), a=a)


def do_darken(value, amount=0.1):
    r, g, b, a = s_rgba(value)

    h, l, s = colorsys.rgb_to_hls(r, g, b)
    l = min(1, max(0, l - s_amount(amount)))

    return rgb_s(*colorsys.hls_to_rgb(h, l, s), a=a)


def do_contrast(value, light, dark, threshold=0.43):
    r, g, b, a = s_rgba(value)

    # Gama correct the values
    r = r / 12.92 if r <= 0.03928 else ((r + 0.055) / 1.055) ** 2.4
    g = g / 12.92 if g <= 0.03928 else ((g + 0.055) / 1.055) ** 2.4
    b = b / 12.92 if b <= 0.03928 else ((b + 0.055) / 1.055) ** 2.4

    luma = 0.2126 * r + 0.7152 * g + 0.0722 * b

    if luma < threshold:
        return light
    else:
        return dark


def do_regex_match(pattern, string, case_insensitive=False):
    flags = 0
    if case_insensitive:
        flags = flags | re.I

    try:
        match = re.match(pattern, string, flags)
    except re.error:
        raise ValueError("Invalid regular expression")
    else:
        results = {}
        if match:
            results.update(
                dict(list(zip(list(range(len(match.groups()))), match.groups())))
            )
            results.update(match.groupdict())

        return results


def do_numberfmt(number, decimal_sep=".", thousand_sep=".", decimal_pos=2, grouping=3):
    """
    Adapted from https://github.com/django/django/blob/main/django/utils/numberformat.py
    Get a number (as a number or string), and return it as a string,
    using formats defined as arguments:

    * decimal_sep: Decimal separator symbol (for example ".")
    * thousand_sep: Thousand separator symbol (for example ",")
    * decimal_pos: Number of decimal positions
    * grouping: Number of digits in every group limited by thousand separator.
        For non-uniform digit grouping, it can be a sequence with the number
        of digit group sizes following the format used by the Python locale
        module in locale.localeconv() LC_NUMERIC grouping (e.g. (3, 2, 0)).
    """
    from decimal import Decimal

    # sign
    sign = ""
    if isinstance(number, Decimal):
        # Format values with more than 200 digits (an arbitrary cutoff) using
        # scientific notation to avoid high memory usage in {:f}'.format().
        _, digits, exponent = number.as_tuple()
        if abs(exponent) + len(digits) > 200:
            number = f"{number:e}"
            coefficient, exponent = number.split("e")
            # Format the coefficient.
            coefficient = do_numberfmt(
                coefficient,
                decimal_sep,
                thousand_sep,
                decimal_pos,
                grouping,
            )
            return f"{coefficient}e{exponent}"
        else:
            str_number = f"{number:f}"
    else:
        str_number = str(number)
    if str_number[0] == "-":
        sign = "-"
        str_number = str_number[1:]

    # decimal part
    if "." in str_number:
        int_part, dec_part = str_number.split(".")
        if decimal_pos is not None:
            dec_part = dec_part[:decimal_pos]
    else:
        int_part, dec_part = str_number, ""
    if decimal_pos is not None:
        dec_part = dec_part + ("0" * (decimal_pos - len(dec_part)))
    dec_part = dec_part and decimal_sep + dec_part

    # grouping
    try:
        # if grouping is a sequence
        intervals = list(grouping)
    except TypeError:
        # grouping is a single value
        intervals = [grouping, 0]
    active_interval = intervals.pop(0)
    int_part_gd = ""
    cnt = 0
    for digit in int_part[::-1]:
        if cnt and cnt == active_interval:
            if intervals:
                active_interval = intervals.pop(0) or active_interval
            int_part_gd += thousand_sep[::-1]
            cnt = 0
        int_part_gd += digit
        cnt += 1
    int_part = int_part_gd[::-1]

    return sign + int_part + dec_part


def do_slugify(value):
    import unicodedata

    value = str(value)
    value = (
        unicodedata.normalize("NFKD", value).encode("ascii", "ignore").decode("ascii")
    )
    value = re.sub(r"[^\w\s-]", "", value).strip().lower()
    return re.sub(r"[-\s]+", "-", value)


def do_qrcode(value):
    from io import BytesIO

    import qrcode
    import qrcode.image.svg

    qr_svg = qrcode.make(value, image_factory=qrcode.image.svg.SvgPathImage)

    svg_stream = BytesIO()
    qr_svg.save(svg_stream)

    svg_value = svg_stream.getvalue().decode("utf-8")

    return Markup(svg_value[svg_value.find("<svg") :])


class LenientJSONEncoder(json.JSONEncoder):
    """
    JSONEncoder subclass that knows how to encode date/time, decimal types, and
    UUIDs.
    """

    def default(self, o):
        # See "Date Time String Format" in the ECMA-262 specification.
        if isinstance(o, datetime.datetime):
            r = o.isoformat()
            if o.microsecond:
                r = r[:23] + r[26:]
            if r.endswith("+00:00"):
                r = r[:-6] + "Z"
            return r
        elif isinstance(o, datetime.date):
            return o.isoformat()
        elif isinstance(o, datetime.time):
            if o.utcoffset() is not None:
                raise ValueError("JSON can't represent timezone-aware times.")
            r = o.isoformat()
            if o.microsecond:
                r = r[:12]
            return r
        elif isinstance(o, datetime.timedelta):
            if o < datetime.timedelta(0):
                sign = "-"
                o *= -1
            else:
                sign = ""

            days = o.days
            seconds = o.seconds
            microseconds = o.microseconds

            minutes = seconds // 60
            seconds = seconds % 60

            hours = minutes // 60
            minutes = minutes % 60
            ms = f".{microseconds:06d}" if microseconds else ""
            return f"{sign}P{days}DT{hours:02d}H{minutes:02d}M{seconds:02d}{ms}S"
        elif isinstance(o, (decimal.Decimal, uuid.UUID)):
            return str(o)
        else:
            return super().default(o)


def do_json(value):
    return Markup(
        json.dumps(value, cls=LenientJSONEncoder)
        .replace("'", r"\u0027")
        .replace("&", r"\u0026")
    )


class LenientUndefined(Undefined):
    """Allows the user to do pretty much anything with undefined values."""

    def _dont_give_up(self, *args, **kwargs):
        return LenientUndefined()

    def __getattr__(self, name):
        if name[:2] == "__":
            raise AttributeError(name)
        return self._dont_give_up()

    def __int__(self):
        return 0

    def __float__(self):
        return 0.0

    __add__ = __radd__ = __mul__ = __rmul__ = __div__ = __rdiv__ = __truediv__ = (
        __rtruediv__
    ) = __floordiv__ = __rfloordiv__ = __mod__ = __rmod__ = __pos__ = __neg__ = (
        __call__
    ) = __getitem__ = __lt__ = __le__ = __gt__ = __ge__ = __complex__ = __pow__ = (
        __rpow__
    ) = _dont_give_up


class ErrorExtension(Extension):
    tags = {"error"}

    def parse(self, parser):
        lineno = next(parser.stream).lineno

        return nodes.Output(
            [self.call_method("_raise", [parser.parse_expression()])]
        ).set_lineno(lineno)

    def _raise(self, message):
        if self.environment.undefined is LenientUndefined:
            return ""

        raise ValueError(message)


def lenient_jinja_env():
    if hasattr(_local, "lenient"):
        return _local.lenient

    lenient = ImmutableSandboxedEnvironment(
        autoescape=True,
        keep_trailing_newline=True,
        undefined=LenientUndefined,
        extensions=[
            ErrorExtension,
            "jinja2.ext.autoescape",
            "jinja2.ext.InternationalizationExtension",
        ],
    )
    lenient.filters = LenientUndefined()
    lenient.install_null_translations(newstyle=True)

    _local.lenient = lenient

    return lenient


class HTMLTemplateParser(HTMLParser):
    def __init__(self):
        self.convert_charrefs = True
        self.reset()
        self._in_title = False
        self.title = None
        self.variables = []
        self.metas = {}
        self.loadsdk_error = False
        self.has_script = False

    def handle_comment(self, data):
        if data == "__loadsdk__" and self.variables:
            self.loadsdk_error = True

    def handle_starttag(self, tag, attrs):
        attrs = dict(attrs)
        self._in_title = tag == "title"

        if tag == "meta" and all(attrs.get(a) for a in ("label", "name", "type")):
            self.variables.append((self.getpos(), attrs))
        elif (
            tag == "meta" and attrs.get("content") and attrs.get("name") in KNOWN_METAS
        ):
            self.metas[attrs["name"]] = attrs["content"]
        elif tag == "script":
            self.has_script = True

    def handle_endtag(self, tag):
        self._in_title = False

    def handle_data(self, data):
        if self._in_title:
            self.title = data.strip()


def extract_app_config(template_text):
    config = {}
    env = lenient_jinja_env()

    template = env.from_string(template_text)

    app_fields = OrderedDict()
    app_attrs = OrderedDict()
    exceptions = []

    def detect_config(
        name,
        type,
        label,
        value=None,
        help=None,
        optional=False,
        *,
        js=False,
        optgroup=None,
        **kwargs,
    ):
        field = {
            "type": type,
            "label": label,
            "value": value,
            "help_text": help,
            "required": not bool(optional),
            "js": js,
            "optgroup": optgroup,
        }

        if js:
            if type not in JS_SUPPORTED_TYPES:
                raise ValueError(
                    f"""
                    "{type}" type of configuration option "{name}" does not support the <code>js</code> parameter.<br>
                    Check the <a href="https://github.com/onsigntv/apps/blob/master/docs/JSBRIDGE.md#app-configuration-object-api">appConfig object documentation</a> to see which types are supported.
                    """
                )

            config.setdefault("js_app_config", []).append(name)

        if type in ("choice", "multichoice"):
            if not isinstance(kwargs.get("choices"), (list, tuple)):
                raise ValueError(
                    f'A "choices" argument is required and must be a list: {name}'
                )

            for choice in kwargs["choices"]:
                if not isinstance(choice, (list, tuple)):
                    raise ValueError(f'Each item in "choices" must be a list: {name}')

                if len(choice) != 2:
                    raise ValueError(
                        f'Each item in "choices" must have two strings (name, label): {name}'
                    )

                if not (isinstance(choice[0], str) and isinstance(choice[1], str)):
                    raise ValueError(
                        f'Each item in "choices" must have two strings (name, label): {name}'
                    )

            if value:
                if value not in [choice[0] for choice in kwargs["choices"]]:
                    raise ValueError(
                        f'Default value must be present in "choices" argument: {name}'
                    )
            else:
                field["value"] = kwargs["choices"][0][0]

            field["choices"] = kwargs["choices"]
        elif type == "font":
            if value and value not in AVAILABLE_FONTS:
                raise ValueError(f"Default font should be a valid system font: {name}")

        app_fields[name] = field

    class DataSinkField:
        def __init__(self, *, name, type, label, help=None, optional=False):
            if not re.match("^[a-zA-Z][a-zA-Z0-9_]*$", name):
                raise ValueError(f"Invalid field name: {name}")

            if type not in DATA_SOURCE_TYPES:
                raise ValueError(f"Data source type invalid: {type}")

            self.name = name
            self.type = type
            self.label = label
            self.help_text = help
            self.required = not bool(optional)

        def __repr__(self):
            return f"<DataSinkField {self.as_dict()}>"

        def as_dict(self):
            return self.__dict__

    def detect_datafeed(
        *, name, label, fields, help=None, optional=False, optgroup=None
    ):
        if not re.match("^[a-zA-Z][a-zA-Z0-9_]*$", name):
            raise ValueError(f"Invalid data feed name: {name}")

        if not isinstance(fields, list):
            raise ValueError("Data feed fields must be a list")

        if not fields:
            raise ValueError("Data feed fields must not be empty")

        if not all(isinstance(f, DataSinkField) for f in fields):
            raise ValueError("Data feed fields must use __field__ constructor")

        field_datafeed = {
            "type": "datafeed",
            "label": label,
            "value": None,
            "help_text": help,
            "required": not bool(optional),
            "datasource_fields": [f.as_dict() for f in fields],
            "optgroup": optgroup,
        }

        app_fields[name] = field_datafeed

    def detect_meta(name, value):
        if name in KNOWN_METAS:
            if name in (
                "audio",
                "audio-app",
                "automation",
                "automation-app",
                "plugin",
                "plugin-app",
            ):
                if value:
                    config[name.split("-")[0]] = True
            else:
                config[name] = value

    def detect_attr(*, name, type, label, mode, help=None, optional=False):
        if not re.match("^[a-zA-Z][a-zA-Z0-9_]*$", name):
            raise ValueError(f"Invalid App Attribute name: {name}")

        if type not in ATTR_TYPES:
            raise ValueError(f"App Attribute type invalid: {type}")

        if mode not in ATTR_MODES:
            raise ValueError(f"App Attribute access mode invalid: {mode}")

        if optional not in {True, False}:
            raise ValueError(
                f'App Attribute optional parameter requires a Boolean value, got: "{optional}"'
            )

        app_attrs[name] = {
            "type": type,
            "label": label,
            "value": True,
            "mode": mode,
            "help_text": help,
            "required": not optional,
        }

    class DetectLoadSDK:
        def __str__(self):
            if app_fields:
                raise ValueError(
                    "{{ __loadsdk__ }} must be added before __config__ or __datafeed__"
                )

            config["sdk"] = True
            return Markup(SDK_TAG)

        def __call__(self, contentinfo=False):
            if contentinfo:
                config["needs_content_info"] = True

            return str(self)

    # Then we render the template to remove all Jinja2 tags
    html = template.render(
        {
            "__loadsdk__": DetectLoadSDK(),
            "__config__": detect_config,
            "__meta__": detect_meta,
            "__datasink__": detect_datafeed,
            "__datafeed__": detect_datafeed,
            "__field__": DataSinkField,
            "__attr__": detect_attr,
        }
    )

    # Now we use a lenient parser to try not to gobble up on invalid markup
    parser = HTMLTemplateParser()
    parser.feed(html)

    if parser.title:
        config["title"] = parser.title
    else:
        raise ValueError("A title tag is required for this app")

    if parser.loadsdk_error:
        raise ValueError("__loadsdk__ must be added before <meta> configurations")

    config["warnings"] = {}

    # After SDK_WARNING_EXPIRY_DATE, not loading sdk will be considered an error.
    if parser.has_script and not config.get("sdk"):
        if datetime.datetime.now().strftime("%Y-%m-%d") >= SDK_WARNING_EXPIRY_DATE:
            raise ValueError("Missing {{ __loadsdk__ }} tag in app.")
        else:
            config["warnings"]["missing_loadsdk"] = True

    config.update(parser.metas)

    if parser.variables:
        if app_fields:
            raise ValueError("Unable to mix __config__ and <meta> variables together.")

        for pos, attrs in parser.variables:
            name = attrs["name"]
            field = {
                "type": attrs["type"],
                "label": attrs["label"],
                "value": attrs.get("value"),
                "help_text": attrs.get("help"),
                "required": "optional" not in attrs,
                "optgroup": attrs.get("optgroup") or None,
            }
            if field["value"] == "":
                field["value"] = None

            if field["type"] in ("choice", "multichoice"):
                if not field.get("value"):
                    raise ValueError(f"Choice variable requires a value: {name}")

                if name not in app_fields:
                    app_fields[name] = field
                    app_fields[name]["choices"] = [[field["value"], field["label"]]]
                elif app_fields[name]["type"] not in ("choice", "multichoice"):
                    raise ValueError(f"Choice variable configured wrong: {name}")
                else:
                    app_fields[name]["choices"].append([field["value"], field["label"]])
            else:
                if field["type"] == "font":
                    if attrs.get("value") and attrs["value"] not in AVAILABLE_FONTS:
                        raise ValueError(
                            f"Default font should be a valid system font: {name}"
                        )
                app_fields[name] = field

        for field in [
            f[1]
            for f in app_fields.items()
            if f[1]["type"] in ("choice", "multichoice")
        ]:
            choices = field["choices"]

            field["label"] = ""
            if all(re.match(r".*\S.*:.*\S.*", choice[1]) for choice in choices):
                field["label"] = choices[0][1].split(":", 1)[0].strip()

                for choice in choices:
                    choice[1] = choice[1].split(":", 1)[1].strip()

    for name, field in app_fields.items():
        if not re.match("^[a-zA-Z][a-zA-Z0-9_]*$", name):
            raise ValueError({"error": f"Invalid variable name: {name}"})

        if name in env.globals or name in BANNED_NAMES:
            raise ValueError({"error": f"Invalid variable name: {name}"})

        if field["type"] not in KNOWN_TYPES | {"datafeed"}:
            raise ValueError({"error": f"Invalid variable type: {name}"})

        if field["type"] in USER_TYPES and field.get("value"):
            raise ValueError(f"User variables cannot contain default values: {name}")

    app_fields["_delay_show"] = {
        "type": "bool",
        "label": "Delay Show Event By 3 Seconds",
        "value": False,
        "help_text": Markup(
            """
            Use this to simulate
            <a href="https://github.com/onsigntv/apps/blob/master/docs/JSBRIDGE.md#show-event" target="_blank">
                app preloading
            </a> done by a real OnSign TV Player.
            """
        ),
        "required": None,
        "optgroup": None,
    }

    app_fields["_proxy_requests"] = {
        "type": "bool",
        "label": "Proxy Javascript Requests",
        "value": False,
        "help_text": Markup(
            """
            Workaround <a href="https://developer.mozilla.org/en-US/docs/Web/HTTP/CORS" target="_blank">
                Cross-Origin Resource Sharing (CORS)
            </a> issues by proxying uses of <code>fetch</code> or <code>XMLHttpRequest</code>
            """
        ),
        "required": None,
        "optgroup": None,
    }

    app_fields["_toast_attr_change"] = {
        "type": "bool",
        "label": "Show Notification on App Attribute Change",
        "value": False,
        "help_text": "Display a toast on the bottom of the screen whenever an app attributes value has changed",
        "required": None,
        "optgroup": None,
    }

    if config.get("plugin"):
        app_fields["_simulate_playback_changes"] = {
            "type": "bool",
            "label": "Simulate Playback Loop Changes",
            "value": False,
            "help_text": Markup(
                """Alternate between two images every 10 seconds, triggering the <a href="https://github.com/onsigntv/apps/blob/master/docs/JSBRIDGE.md#signage-playbackloopschanged-event" target="_blank">
                <code>playbackloopschanged</code>
                </a> event."""
            ),
            "required": None,
            "optgroup": None,
        }

    app_fields["_serial_port_config"] = {
        "type": "text",
        "label": "Serial Port Configuration",
        "value": "",
        "help_text": "",
        "required": None,
        "optgroup": None,
    }

    app_fields["_toast_serial_port_data"] = {
        "type": "bool",
        "label": "Show Notification When Serial Port Data is Read",
        "value": False,
        "help_text": Markup(
            """
            Display a toast on the bottom of the screen whenever the
            <a href="https://github.com/onsigntv/apps/blob/master/docs/JSBRIDGE.md#signage-serialportdata-event">
            <code>serialportdata</code> event</a> is fired
            """
        ),
        "required": None,
        "optgroup": None,
    }

    app_fields["_playback_info"] = {
        "type": "paragraph",
        "label": "Javascript Playback Info Data",
        "value": samples.PLAYBACK_INFO,
        "help_text": Markup(
            """
            Check the
            <a href="https://github.com/onsigntv/apps/blob/master/docs/JSBRIDGE.md" target="_blank">
              Javascript API documentation
            </a>
            to know what information will be available through
            <a href="https://github.com/onsigntv/apps/blob/master/docs/JSBRIDGE.md#playbackInfo" target="_blank">
              <code>signage.playbackInfo()</code>
            </a>
            """
        ),
        "required": None,
        "optgroup": None,
    }

    config["fields"] = list(map(list, app_fields.items()))
    config["attrs"] = list(map(list, app_attrs.items()))
    config["exceptions"] = exceptions

    try:
        node = env.parse(template_text)
        list(env.lex(env.preprocess(template_text)))
    except TemplateSyntaxError:
        pass
    else:
        catalog = OrderedDict()

        for lineno, func, message in extract_from_ast(
            node, GETTEXT_FUNCTIONS, babel_style=False
        ):
            if message not in catalog:
                catalog[message] = {"msgid": message}

        if catalog:
            config["i18n"] = True
            config["catalog"] = list(catalog.values())

    return config
